@@ -3,8 +3,10 @@
 import datetime as dt
 import json
 import logging
+import os
 import select
 import threading
+import time
 from collections import OrderedDict, defaultdict
 from datetime import datetime
 from weakref import WeakKeyDictionary
@@ -15,10 +17,11 @@ from sql.functions import CurrentTimestamp, Function
 
 from trytond import backend
 from trytond.config import config
+from trytond.pool import Pool
 from trytond.transaction import Transaction
 from trytond.tools import resolve, grouped_slice
 
-__all__ = ['BaseCache', 'Cache', 'LRUDict']
+__all__ = ['BaseCache', 'Cache', 'LRUDict', 'LRUDictTransaction']
 _clear_timeout = config.getint('cache', 'clean_timeout', default=5 * 60)
 logger = logging.getLogger(__name__)
 
@@ -28,7 +31,7 @@ def _cast(column):
         __slots__ = ()
         _function = 'DATETIME'
 
-    if backend.name() == 'sqlite':
+    if backend.name == 'sqlite':
         column = SQLite_DateTime(column)
     return column
 
@@ -43,6 +46,22 @@ def freeze(o):
         return o
 
 
+def unfreeze(o):
+    if isinstance(o, frozenset):
+        return dict((x, unfreeze(y)) for x, y in o)
+    else:
+        return o
+
+
+def _get_modules(cursor):
+    ir_module = Table('ir_module')
+    cursor.execute(*ir_module.select(
+            ir_module.name,
+            where=ir_module.state.in_(
+                ['activated', 'to upgrade', 'to remove'])))
+    return {m for m, in cursor}
+
+
 class BaseCache(object):
     _instances = {}
 
@@ -50,6 +69,7 @@ class BaseCache(object):
         self._name = name
         self.size_limit = size_limit
         self.context = context
+        self.hit = self.miss = 0
         if isinstance(duration, dt.timedelta):
             self.duration = duration
         elif isinstance(duration, (int, float)):
@@ -61,11 +81,19 @@ class BaseCache(object):
         assert self._name not in self._instances
         self._instances[self._name] = self
 
+    @classmethod
+    def stats(cls):
+        for name, inst in cls._instances.items():
+            yield {
+                'name': name,
+                'hit': inst.hit,
+                'miss': inst.miss,
+                }
+
     def _key(self, key):
         if self.context:
             context = Transaction().context.copy()
             context.pop('client', None)
-            context.pop('_check_access', None)
             context.pop('_request', None)
             return (key, Transaction().user, freeze(context))
         return key
@@ -79,10 +107,18 @@ class BaseCache(object):
     def clear(self):
         raise NotImplementedError
 
+    @classmethod
+    def clear_all(cls):
+        for inst in cls._instances.values():
+            inst.clear()
+
     @classmethod
     def sync(cls, transaction):
         raise NotImplementedError
 
+    def sync_since(self, value):
+        raise NotImplementedError
+
     @classmethod
     def commit(cls, transaction):
         raise NotImplementedError
@@ -104,7 +140,7 @@ class MemoryCache(BaseCache):
     _clean_last = datetime.now()
     _default_lower = Transaction.monotonic_time()
     _listener = {}
-    _listener_lock = threading.Lock()
+    _listener_lock = defaultdict(threading.Lock)
     _table = 'ir_cache'
     _channel = _table
 
@@ -136,10 +172,13 @@ class MemoryCache(BaseCache):
         try:
             (expire, result) = cache.pop(key)
             if expire and expire < dt.datetime.now():
+                self.miss += 1
                 return default
             cache[key] = (expire, result)
+            self.hit += 1
             return result
         except (KeyError, TypeError):
+            self.miss += 1
             return default
 
     def set(self, key, value):
@@ -173,11 +212,15 @@ class MemoryCache(BaseCache):
         database = transaction.database
         dbname = database.name
         if not _clear_timeout and database.has_channel():
-            with cls._listener_lock:
-                if dbname not in cls._listener:
-                    cls._listener[dbname] = listener = threading.Thread(
+            pid = os.getpid()
+            with cls._listener_lock[pid]:
+                if (pid, dbname) not in cls._listener:
+                    cls._listener[pid, dbname] = listener = threading.Thread(
                         target=cls._listen, args=(dbname,), daemon=True)
                     listener.start()
+                    while (not getattr(listener, 'listening', False)
+                            and listener.is_alive()):
+                        time.sleep(.01)
             return
         if (datetime.now() - cls._clean_last).total_seconds() < _clear_timeout:
             return
@@ -188,8 +231,9 @@ class MemoryCache(BaseCache):
                 cursor.execute(*table.select(
                         _cast(table.timestamp), table.name))
                 timestamps = {}
-                for timestamp, name in cursor.fetchall():
+                for timestamp, name in cursor:
                     timestamps[name] = timestamp
+                modules = _get_modules(cursor)
         finally:
             database.put_connection(connection)
         for name, timestamp in timestamps.items():
@@ -200,8 +244,12 @@ class MemoryCache(BaseCache):
             inst_timestamp = inst._timestamp.get(dbname)
             if not inst_timestamp or timestamp > inst_timestamp:
                 inst._clear(dbname, timestamp)
+        Pool(dbname).refresh(modules)
         cls._clean_last = datetime.now()
 
+    def sync_since(self, value):
+        return self._clean_last > value
+
     @classmethod
     def commit(cls, transaction):
         table = Table(cls._table)
@@ -253,6 +301,7 @@ class MemoryCache(BaseCache):
                 connection.commit()
             finally:
                 database.put_connection(connection)
+            cls._clean_last = datetime.now()
         reset.clear()
 
     @classmethod
@@ -264,11 +313,11 @@ class MemoryCache(BaseCache):
 
     @classmethod
     def drop(cls, dbname):
-        with cls._listener_lock:
-            listener = cls._listener.pop(dbname, None)
+        pid = os.getpid()
+        with cls._listener_lock[pid]:
+            listener = cls._listener.pop((pid, dbname), None)
         if listener:
-            Database = backend.get('Database')
-            database = Database(dbname)
+            database = backend.Database(dbname)
             conn = database.get_connection()
             try:
                 cursor = conn.cursor()
@@ -282,42 +331,61 @@ class MemoryCache(BaseCache):
             inst._database_cache.pop(dbname, None)
             inst._transaction_lower.pop(dbname, None)
 
+    @classmethod
+    def refresh_pool(cls, transaction):
+        database = transaction.database
+        dbname = database.name
+        if not _clear_timeout and database.has_channel():
+            database = backend.Database(dbname)
+            conn = database.get_connection()
+            try:
+                cursor = conn.cursor()
+                cursor.execute(
+                    'NOTIFY "%s", %%s' % cls._channel, ('refresh pool',))
+                conn.commit()
+            finally:
+                database.put_connection(conn)
+
     @classmethod
     def _listen(cls, dbname):
-        Database = backend.get('Database')
-        database = Database(dbname)
+        database = backend.Database(dbname)
         if not database.has_channel():
             raise NotImplementedError
 
         logger.info("listening on channel '%s' of '%s'", cls._channel, dbname)
-        conn = database.get_connection()
+        conn = database.get_connection(autocommit=True)
+        pid = os.getpid()
+        current_thread = threading.current_thread()
         try:
             cursor = conn.cursor()
             cursor.execute('LISTEN "%s"' % cls._channel)
-            conn.commit()
+            current_thread.listening = True
 
-            while cls._listener.get(dbname) == threading.current_thread():
-                readable, _, _ = select.select([conn], [], [])
+            while cls._listener.get((pid, dbname)) == current_thread:
+                readable, _, _ = select.select([conn], [], [], 60)
                 if not readable:
                     continue
 
                 conn.poll()
                 while conn.notifies:
                     notification = conn.notifies.pop()
-                    if notification.payload:
+                    if notification.payload == 'refresh pool':
+                        Pool(dbname).refresh(_get_modules(cursor))
+                    elif notification.payload:
                         reset = json.loads(notification.payload)
                         for name in reset:
                             inst = cls._instances[name]
                             inst._clear(dbname)
+                cls._clean_last = datetime.now()
         except Exception:
             logger.error(
                 "cache listener on '%s' crashed", dbname, exc_info=True)
             raise
         finally:
             database.put_connection(conn)
-            with cls._listener_lock:
-                if cls._listener.get(dbname) == threading.current_thread():
-                    del cls._listener[dbname]
+            with cls._listener_lock[pid]:
+                if cls._listener.get((pid, dbname)) == current_thread:
+                    del cls._listener[pid, dbname]
 
 
 if config.get('cache', 'class'):
@@ -330,19 +398,28 @@ class LRUDict(OrderedDict):
     """
     Dictionary with a size limit.
     If size limit is reached, it will remove the first added items.
+    The default_factory provides the same behavior as in standard
+    collections.defaultdict.
     """
     __slots__ = ('size_limit',)
 
-    def __init__(self, size_limit, *args, **kwargs):
+    def __init__(self, size_limit, default_factory=None, *args, **kwargs):
         assert size_limit > 0
         self.size_limit = size_limit
         super(LRUDict, self).__init__(*args, **kwargs)
+        self.default_factory = default_factory
         self._check_size_limit()
 
     def __setitem__(self, key, value):
         super(LRUDict, self).__setitem__(key, value)
         self._check_size_limit()
 
+    def __missing__(self, key):
+        if self.default_factory is None:
+            raise KeyError(key)
+        self[key] = value = self.default_factory()
+        return value
+
     def update(self, *args, **kwargs):
         super(LRUDict, self).update(*args, **kwargs)
         self._check_size_limit()
@@ -359,7 +436,7 @@ class LRUDict(OrderedDict):
 
 class LRUDictTransaction(LRUDict):
     """
-    Dictionary with a size limit. (see LRUDict)
+    Dictionary with a size limit and default_factory. (see LRUDict)
     It is refreshed when transaction counter is changed.
     """
     __slots__ = ('transaction', 'counter')
