@@ -2,7 +2,6 @@
 # this repository contains the full copyright notices and license terms.
 import logging
 import os
-import xml.dom.minidom
 from difflib import SequenceMatcher
 from collections import defaultdict
 from hashlib import md5
@@ -15,31 +14,25 @@ from sql.functions import Substring, Position
 from sql.conditionals import Case
 from sql.aggregate import Max
 
+from defusedxml.minidom import parseString
 from genshi.filters.i18n import extract as genshi_extract
 from relatorio.reporting import MIMETemplateLoader
 from relatorio.templates.opendocument import get_zip_file
 
+from trytond.cache import Cache
+from trytond.config import config
 from trytond.exceptions import UserError
 from trytond.i18n import gettext
-from ..model import ModelView, ModelSQL, fields
-from ..wizard import Wizard, StateView, StateTransition, StateAction, \
-    Button
-from ..tools import file_open, grouped_slice, cursor_dict
-from ..pyson import PYSONEncoder, Eval
-from ..transaction import Transaction
-from ..pool import Pool
-from ..cache import Cache
-from ..config import config
+from trytond.model import ModelView, ModelSQL, fields
+from trytond.pool import Pool
+from trytond.pyson import PYSONEncoder, Eval
+from trytond.tools import file_open, grouped_slice, cursor_dict
+from trytond.tools.string_ import LazyString, StringPartitioned
+from trytond.transaction import Transaction
+from trytond.wizard import (
+    Wizard, StateView, StateTransition, StateAction, Button)
 
-from trytond.ir.lang import get_parent_language as get_parent
-
-__all__ = ['Translation',
-    'TranslationSetStart', 'TranslationSetSucceed', 'TranslationSet',
-    'TranslationCleanStart', 'TranslationCleanSucceed', 'TranslationClean',
-    'TranslationUpdateStart', 'TranslationUpdate',
-    'TranslationExportStart', 'TranslationExportResult', 'TranslationExport',
-    'TranslationReport',
-    ]
+from .lang import get_parent_language as get_parent
 
 logger = logging.getLogger(__name__)
 
@@ -83,6 +76,8 @@ class Translation(ModelSQL, ModelView):
     overriding_module = fields.Char('Overriding Module', readonly=True)
     _translation_cache = Cache('ir.translation', size_limit=10240,
         context=False)
+    _translation_report_cache = Cache(
+        'ir.translation.get_report', context=False)
     _get_language_cache = Cache('ir.translation.get_language')
 
     @classmethod
@@ -179,6 +174,8 @@ class Translation(ModelSQL, ModelView):
             for val in string:
                 if not val or val in translations[type][name]:
                     continue
+                if isinstance(val, LazyString):
+                    continue
                 cursor.execute(
                     *ir_translation.insert(columns,
                         [[name, 'en', type, val, '', module_name, False, -1]]))
@@ -217,7 +214,8 @@ class Translation(ModelSQL, ModelView):
                             ir_translation.type, ir_translation.src,
                             ir_translation.value, ir_translation.module,
                             ir_translation.fuzzy, ir_translation.res_id],
-                        [[trans_name, 'en',
+                        [[
+                                trans_name, 'en',
                                 'wizard_button', button.string,
                                 '', module_name,
                                 False, -1]]))
@@ -225,8 +223,8 @@ class Translation(ModelSQL, ModelView):
                 cursor.execute(*ir_translation.update(
                         [ir_translation.src],
                         [button.string],
-                        where=ir_translation.id ==
-                        trans_buttons[trans_name]['id']))
+                        where=ir_translation.id
+                        == trans_buttons[trans_name]['id']))
 
         for state_name, state in wizard.states.items():
             if not isinstance(state, StateView):
@@ -265,7 +263,8 @@ class Translation(ModelSQL, ModelView):
         Operator = fields.SQL_OPERATORS[operator]
         return [('id', 'in', table.select(table.id,
                     where=Operator(Substring(table.name, 1,
-                            Case((Position(',', table.name) > 0,
+                            Case((
+                                    Position(',', table.name) > 0,
                                     Position(',', table.name) - 1),
                                 else_=0)), value)))]
 
@@ -286,11 +285,13 @@ class Translation(ModelSQL, ModelView):
         return [('/form//field[@name="value"]', 'spell', Eval('lang'))]
 
     @classmethod
-    def get_ids(cls, name, ttype, lang, ids):
+    def get_ids(cls, name, ttype, lang, ids, cached_after=None):
         "Return translation for each id"
         pool = Pool()
         ModelFields = pool.get('ir.model.field')
         Model = pool.get('ir.model')
+        context = Transaction().context
+        fuzzy_translation = context.get('fuzzy_translation', False)
 
         translations, to_fetch = {}, []
         name = str(name)
@@ -324,11 +325,38 @@ class Translation(ModelSQL, ModelView):
                 else:
                     name = record.model + ',' + field_name
                 translations[record.id] = cls.get_source(name, ttype, lang)
+                if translations[record.id] is None:
+                    with Transaction().set_context(language=lang):
+                        if ttype in {'field', 'help'}:
+                            try:
+                                field = getattr(
+                                    pool.get(record.model.model), record.name)
+                            except KeyError:
+                                continue
+                            translations[record.id] = ''
+                            if ttype == 'field':
+                                value = field.string
+                            else:
+                                value = field.help
+                        else:
+                            try:
+                                model = pool.get(record.model)
+                            except KeyError:
+                                continue
+                            if not model.__doc__:
+                                continue
+                            value = model._get_name()
+                        if isinstance(value, StringPartitioned):
+                            for source in value:
+                                translations[record.id] += source
+                        else:
+                            translations[record.id] = value
             return translations
 
         # Don't use cache for fuzzy translation
-        if not Transaction().context.get(
-                'fuzzy_translation', False):
+        if (not fuzzy_translation
+                and (not cached_after
+                    or not cls._translation_cache.sync_since(cached_after))):
             for obj_id in ids:
                 trans = cls._translation_cache.get((name, ttype, lang, obj_id),
                     -1)
@@ -346,7 +374,7 @@ class Translation(ModelSQL, ModelView):
                 translations.update(
                     cls.get_ids(name, ttype, parent_lang, to_fetch))
 
-            if Transaction().context.get('fuzzy_translation', False):
+            if fuzzy_translation:
                 fuzzy_clause = []
             else:
                 fuzzy_clause = [('fuzzy', '=', False)]
@@ -361,7 +389,7 @@ class Translation(ModelSQL, ModelView):
                             ] + fuzzy_clause):
                     translations[translation.res_id] = translation.value
             # Don't store fuzzy translation in cache
-            if not Transaction().context.get('fuzzy_translation', False):
+            if not fuzzy_translation:
                 for res_id in to_fetch:
                     value = translations.setdefault(res_id)
                     cls._translation_cache.set(
@@ -579,6 +607,29 @@ class Translation(ModelSQL, ModelView):
             cls._translation_cache.set(key, res[key])
         return res
 
+    @classmethod
+    def get_report(cls, report_name, text):
+        language = Transaction().language
+        key = (report_name, language)
+        if cls._translation_report_cache.get(key) is None:
+            cache = {}
+            code = language
+            while code:
+                translations = cls.search([
+                        ('lang', '=', code),
+                        ('type', '=', 'report'),
+                        ('name', '=', report_name),
+                        ('value', '!=', ''),
+                        ('value', '!=', None),
+                        ('fuzzy', '=', False),
+                        ('res_id', '=', -1),
+                        ], order=[('module', 'DESC')])
+                for translation in translations:
+                    cache.setdefault(translation.src, translation.value)
+                code = get_parent(code)
+            cls._translation_report_cache.set(key, cache)
+        return cls._translation_report_cache.get(key, {}).get(text, text)
+
     @classmethod
     def delete(cls, translations):
         pool = Pool()
@@ -586,9 +637,10 @@ class Translation(ModelSQL, ModelView):
         Model = pool.get('ir.model')
         ModelField = pool.get('ir.model.field')
         Message._message_cache.clear()
-        Model._get_name_cache.clear()
+        Model._get_names_cache.clear()
         ModelField._get_name_cache.clear()
         cls._translation_cache.clear()
+        cls._translation_report_cache.clear()
         ModelView._fields_view_get_cache.clear()
         return super(Translation, cls).delete(translations)
 
@@ -599,9 +651,10 @@ class Translation(ModelSQL, ModelView):
         Model = pool.get('ir.model')
         ModelField = pool.get('ir.model.field')
         Message._message_cache.clear()
-        Model._get_name_cache.clear()
+        Model._get_names_cache.clear()
         ModelField._get_name_cache.clear()
         cls._translation_cache.clear()
+        cls._translation_report_cache.clear()
         ModelView._fields_view_get_cache.clear()
         vlist = [x.copy() for x in vlist]
 
@@ -618,9 +671,10 @@ class Translation(ModelSQL, ModelView):
         Model = pool.get('ir.model')
         ModelField = pool.get('ir.model.field')
         Message._message_cache.clear()
-        Model._get_name_cache.clear()
+        Model._get_names_cache.clear()
         ModelField._get_name_cache.clear()
         cls._translation_cache.clear()
+        cls._translation_report_cache.clear()
         ModelView._fields_view_get_cache.clear()
         return super(Translation, cls).write(*args)
 
@@ -721,7 +775,6 @@ class Translation(ModelSQL, ModelView):
                         'report', 'view', 'wizard_button', 'selection'}:
                     domain.append(('src', '=', new_translation.src))
 
-
                 # Translation updates with overrides have always to be run
                 # twice after the initial import, because the original entry is
                 # not yet saved and thus cannot be found. Also we must take care
@@ -936,7 +989,7 @@ class TranslationSet(Wizard):
                 zip_.read('content.xml'),
                 zip_.read('styles.xml'),
                 ]:
-            document = xml.dom.minidom.parseString(content_xml)
+            document = parseString(content_xml)
             for string in extract(document.documentElement):
                 yield string
 
@@ -976,11 +1029,10 @@ class TranslationSet(Wizard):
         pool = Pool()
         Report = pool.get('ir.action.report')
         Translation = pool.get('ir.translation')
-        context = Transaction().context
 
-        if context.get('active_model') == Report.__name__:
-            reports = Report.browse(context.get('active_ids', []))
-        elif context.get('active_model', 'ir.ui.menu') == 'ir.ui.menu':
+        if self.model == Report:
+            reports = self.records
+        elif not self.model or self.model.__name__ == 'ir.ui.menu':
             with Transaction().set_context(active_test=False):
                 reports = Report.search([('translatable', '=', True)])
         else:
@@ -1044,7 +1096,8 @@ class TranslationSet(Wizard):
                                     translation.type, translation.src,
                                     translation.value, translation.module,
                                     translation.fuzzy, translation.res_id],
-                                [[report.report_name, 'en',
+                                [[
+                                        report.report_name, 'en',
                                         'report', string,
                                         '', module,
                                         False, -1]]))
@@ -1070,11 +1123,10 @@ class TranslationSet(Wizard):
         pool = Pool()
         View = pool.get('ir.ui.view')
         Translation = pool.get('ir.translation')
-        context = Transaction().context
 
-        if context.get('active_model') == View.__name__:
-            views = View.browse(context.get('active_ids', []))
-        elif context.get('active_model', 'ir.ui.menu') == 'ir.ui.menu':
+        if self.model == View:
+            views = self.records
+        elif not self.model or self.model.__name__ == 'ir.ui.menu':
             with Transaction().set_context(active_test=False):
                 views = View.search([])
         else:
@@ -1131,8 +1183,8 @@ class TranslationSet(Wizard):
                                 [translation.src,
                                     translation.fuzzy],
                                 [string, True],
-                                where=(translation.id ==
-                                    trans_views[string_trans]['id'])))
+                                where=(translation.id
+                                    == trans_views[string_trans]['id'])))
                         del trans_views[string_trans]
                         done = True
                         break
@@ -1142,7 +1194,8 @@ class TranslationSet(Wizard):
                                 translation.type, translation.src,
                                 translation.value, translation.module,
                                 translation.fuzzy, translation.res_id],
-                            [[view.model, 'en',
+                            [[
+                                    view.model, 'en',
                                     'view', string,
                                     '', view.module,
                                     False, -1]]))
@@ -1195,7 +1248,10 @@ class TranslationClean(Wizard):
             Model = pool.get(model_name)
         except KeyError:
             return True
-        if field_name not in Model._fields:
+        field = Model._fields.get(field_name)
+        if not field:
+            return True
+        if translation.src not in list(field.string):
             return True
 
     @staticmethod
@@ -1293,10 +1349,13 @@ class TranslationClean(Wizard):
             Model = pool.get(model_name)
         except KeyError:
             return True
-        if field_name not in Model._fields:
+        field = Model._fields.get(field_name)
+        if not field:
+            return True
+        if not field.help:
+            return True
+        if translation.src not in list(field.help):
             return True
-        field = Model._fields[field_name]
-        return not field.help
 
     def transition_clean(self):
         pool = Pool()
@@ -1375,19 +1434,18 @@ class TranslationUpdate(Wizard):
         Translation = pool.get('ir.translation')
         Report = pool.get('ir.action.report')
         View = pool.get('ir.ui.view')
-        context = Transaction().context
         cursor = Transaction().connection.cursor()
         cursor_update = Transaction().connection.cursor()
         translation = Translation.__table__()
         lang = self.start.language.code
         parent_lang = get_parent(lang)
 
-        if context.get('active_model') == Report.__name__:
-            reports = Report.browse(context.get('active_ids', []))
+        if self.model == Report:
+            reports = self.records
             source_clause = ((translation.type == 'report')
                 & translation.name.in_([r.report_name for r in reports]))
-        elif context.get('active_model') == View.__name__:
-            views = View.browse(context.get('active_ids', []))
+        elif self.model == View:
+            views = self.records
             source_clause = ((translation.type == 'view')
                 & translation.name.in_([v.model for v in views]))
         else:
@@ -1438,7 +1496,7 @@ class TranslationUpdate(Wizard):
                         & (translation.module == row['module'])
                         & (translation.lang == lang)))
 
-        if context.get('active_model') in {Report.__name__, View.__name__}:
+        if self.model in {Report, View}:
             return
 
         columns = [translation.name.as_('name'),
@@ -1569,7 +1627,10 @@ class TranslationExportResult(ModelView):
     "Export translation"
     __name__ = 'ir.translation.export.result'
 
-    file = fields.Binary('File', readonly=True)
+    language = fields.Many2One('ir.lang', 'Language', readonly=True)
+    module = fields.Many2One('ir.module', 'Module', readonly=True)
+    file = fields.Binary('File', readonly=True, filename='filename')
+    filename = fields.Char('Filename')
 
 
 class TranslationExport(Wizard):
@@ -1600,7 +1661,10 @@ class TranslationExport(Wizard):
         cast = self.result.__class__.file.cast
         self.result.file = False  # No need to store it in session
         return {
+            'module': self.start.module.id,
+            'language': self.start.language.id,
             'file': cast(file_) if file_ else None,
+            'filename': '%s.po' % self.start.language.code,
             }
 
 
@@ -1611,11 +1675,8 @@ class TranslationReport(Wizard):
     open_ = StateAction('ir.act_translation_report')
 
     def do_open_(self, action):
-        pool = Pool()
-        Report = pool.get('ir.action.report')
         context = Transaction().context
-        assert context['active_model'] == Report.__name__
-        reports = Report.browse(context['active_ids'])
+        reports = self.records
         action['pyson_domain'] = PYSONEncoder().encode([
                 ('type', '=', 'report'),
                 ('name', 'in', [r.report_name for r in reports]),
